
/*
 * N.B.
 *
 * This version of Sam Smith's unzip code is hacked to run on the Atari ST
 * no probs, as far as i have tested it.
 *
 * This source is written for the Laser C compiler, and i've added assembler
 * routines in places in the style eaten by the Laser compiler.  These
 * assembler routines are in the two byte-swapping functions, and more
 * usefully in the FillBitBuffer function and the CRC calculation function.
 * i've also named lots of variables as 'register' since the 68000 has more
 * registers than the 80x86 family.
 *
 * In addition to that, i've included two needed standard functions that
 * the Laser C package lacks, memcpy and strchr.  i wrote the body of
 * memcpy in assembler to speed it up a bit.
 *
 * On top of that i've rewritten the UnImplode code to generate and use
 * binary trees rather than the original Shannon-Fano trees.  This speeds
 * UnImplode by about 2X with my setup, which uses a ramdisk that
 * compresses the files stored in it, so it's about the speed of a
 * floppy drive when it's written to.
 *
 * Here i reiterate Mr.Smith's notice:  do not use any of this in a
 * commercial product.  If you use this in your own program you Must
 * distribute source code.
 *
 * 
 * i am not completely certain of the copyrights issues involved, but i,
 * ben soo, reserve the rights to the modifications i made to this code.
 *
 * ben soo, Sept.1990
 *
 */

/*
 * Copyright 1989 Samuel H. Smith;  All rights reserved
 *
 * Do not distribute modified versions without my permission.
 * Do not remove or alter this notice or any other copyright notice.
 * If you use this in your own program you must distribute source code.
 * Do not use any of this in a commercial product.
 *
 */

/*
 * UnZip - A simple zipfile extract utility
 *
 * To compile:
 *      compile unzip.c, crc32.c, memcpy.c, and strchr.c, and link
 *      them together.
 *
 */

#define VERSION  "UnZip:  Zipfile Extract v2.01 (C) of 09-09-89;  (C) 1989 Samuel H. Smith\n        (modified for the ST, and to add some speed, by ben soo 1990)"

typedef unsigned char byte;    /* code assumes UNSIGNED bytes */
typedef long longint;
typedef unsigned int word;
typedef char boolean;

#define STRSIZ 256

#include <stdio.h>
 /* this is your standard header for all C compiles */

#include <define.h>
#include <ctype.h>
#include <osbind.h>



/*
 * SEE HOST OPERATING SYSTEM SPECIFICS SECTION STARTING NEAR LINE 180
 *
 */


/* ----------------------------------------------------------- */
/*
 * Zipfile layout declarations
 *
 */

typedef longint signature_type;


#define LOCAL_FILE_HEADER_SIGNATURE  0x04034b50L


typedef struct local_file_header {
    word version_needed_to_extract;
    word general_purpose_bit_flag;
    word compression_method;
    word last_mod_file_time;
    word last_mod_file_date;
    longint crc32;
    longint compressed_size;
    longint uncompressed_size;
    word filename_length;
    word extra_field_length;
} local_file_header;


#define CENTRAL_FILE_HEADER_SIGNATURE  0x02014b50L


typedef struct central_directory_file_header {
    word version_made_by;
    word version_needed_to_extract;
    word general_purpose_bit_flag;
    word compression_method;
    word last_mod_file_time;
    word last_mod_file_date;
    longint crc32;
    longint compressed_size;
    longint uncompressed_size;
    word filename_length;
    word extra_field_length;
    word file_comment_length;
    word disk_number_start;
    word internal_file_attributes;
    longint external_file_attributes;
    longint relative_offset_local_header;
} central_directory_file_header;


#define END_CENTRAL_DIR_SIGNATURE  0x06054b50L


typedef struct end_central_dir_record {
    word number_this_disk;
    word number_disk_with_start_central_directory;
    word total_entries_central_dir_on_this_disk;
    word total_entries_central_dir;
    longint size_central_directory;
    longint offset_start_central_directory;
    word zipfile_comment_length;
} end_central_dir_record;



/* ----------------------------------------------------------- */
/*
 * input file variables
 *
 */

#define INBUFSIZ 0x6000
byte *inbuf;            /* input file buffer - any size is legal */
byte *inptr;

long incnt;
unsigned bitbuf;
int bits_left;
boolean zipeof;

int zipfd;
char zipfn[STRSIZ];
local_file_header lrec;


/* ----------------------------------------------------------- */
/*
 * output stream variables
 *
 */

#define OUTBUFSIZ 0x6000        /* must be 0x2000 or larger for unImplode */
byte *outbuf;                   /* buffer for rle look-back */
byte *outptr;

longint outpos;            /* absolute position in outfile */
long outcnt;            /* current position in outbuf */

int outfd;
char filename[STRSIZ];
char extra[STRSIZ];

#define DLE 144


/* ----------------------------------------------------------- */
/*
 * shrink/reduce working storage
 *
 */

int factor;
byte followers[256][64];
byte Slen[256];

#define max_bits 13
#define init_bits 9
#define hsize 8192
#define first_ent 257
#define clear 256

typedef int hsize_array_integer[hsize+1];
typedef byte hsize_array_byte[hsize+1];

hsize_array_integer prefix_of;
hsize_array_byte suffix_of;
hsize_array_byte stack;

int codesize;
int maxcode;
int free_ent;
int maxcodemax;
int offset;
int sizex;



/* ============================================================= */
/*
 * Host operating system details
 *
 */

#include <strings.h>
 /* this include defines strcpy, strcmp, etc. */

#include <stdio.h>
#define SEEK_CUR  1
#define SEEK_END  2
#define SEEK_SET  0

#include <fcntl.h>
#ifdef HIGH_LOW
#undef HIGH_LOW
#else
#define HIGH_LOW
#endif
 /*
  * change 'undef' to 'define' if your machine stores high order bytes in
  * lower addresses.
  */

void set_file_time()
 /*
  * set the output file date/time stamp according to information from the
  * zipfile directory record for this file 
  */
{
    struct {
                        word ztime;     /* date and time words */
                        word zdate;     /* .. same format as in .ZIP file */
            } td;

    /*
     * set output file date and time - this is optional and can be
     * deleted if your compiler does not easily support setftime() 
     */

    td.ztime = lrec.last_mod_file_time;
    td.zdate = lrec.last_mod_file_date;

/* Remember, any header field we want to fiddle with we first have to
   swap into 68000 memory format. */

    Fdatime(&td, outfd, 1);
}


int create_output_file()
 /* return non-0 if creat failed */
{

    /* create the output file with READ and WRITE permissions */
    outfd = Fcreate(filename, 0);
    if (outfd < 1) {
        printf("Can't create output: %s\n", filename);
        return 1;
    }

    /*
     * close the newly created file and reopen it in BINARY mode to
     * disable all CR/LF translations
     *
     * On the ST, Fopen does only binary, so we reopen it write-only.
     */
    close(outfd);
    outfd = Fopen(filename, 1);

    /* write a single byte at EOF to pre-allocate the file */
        Fseek(lrec.uncompressed_size - 1L, outfd, SEEK_SET);
    Fwrite(outfd, 1L, "?");
    Fseek(0L, outfd, SEEK_SET);
    return 0;
}


int open_input_file()
 /* return non-0 if creat failed */
{
    /*
     * open the zipfile for reading and in BINARY mode to prevent cr/lf
     * translation, which would corrupt the bitstreams
     *
     * On the ST, Fopen does binary anyhow, so we open it read-only.
     */

    zipfd = Fopen(zipfn, 0);
    if (zipfd < 1) {
        printf("Can't open input file: %s\n", zipfn);
        return (1);
    }
    return 0;
}


#ifdef HIGH_LOW

void swap_bytes(wordp)
register word *wordp;

 /* convert intel style 'short int' variable to host format */
{
register swapw = *wordp;

asm
    {
    rol.w #8,swapw
    }
    *wordp = swapw;
}

void swap_lbytes(longp)
register longint *longp;

 /* convert intel style 'long' variable to host format */
{
    register longint charp0 = *longp;
    register char cnt = 16;

/*  rol.w rol.w rol.w your boat, gently down the stream.. */
asm
    {
    rol.w #8,charp0;
    rol.l cnt,charp0;
    rol.w #8, charp0;
    }
    *longp = charp0;
}

#endif



/* ============================================================= */

int FillBuffer()
 /* fill input buffer if possible */
{
    long readsize;

        if (lrec.compressed_size <= 0)
        return incnt = 0;

        if (lrec.compressed_size > INBUFSIZ)
        readsize = INBUFSIZ;
    else
                readsize =  lrec.compressed_size;
    incnt = Fread(zipfd, readsize, inbuf,);

        lrec.compressed_size -= incnt;
    inptr = inbuf;
    return incnt--;
}

int ReadByte(x)
unsigned *x;
 /* read a byte; return 8 if byte available, 0 if not */
{
    if (incnt-- == 0)
        if (FillBuffer() == 0)
            return 0;

    *x = *inptr++;
    return 8;
}

/*
int ReadWord(x)
unsigned *x;
  read a word; return 16
{
    register word z;
    register word *winptr = inptr;
    z = *winptr++;
    asm
    {
    rol.w    #8,    z;
    }
    *x = z;
    inptr = winptr;
    incnt -= 2;
    return 16;
}
*/

/* ------------------------------------------------------------- */
static unsigned mask_bits[] =
        {0,     0x0001, 0x0003, 0x0007, 0x000f,
                0x001f, 0x003f, 0x007f, 0x00ff,
                0x01ff, 0x03ff, 0x07ff, 0x0fff,
                0x1fff, 0x3fff, 0x7fff, 0xffff
        };


int FillBitBuffer(bits)
register int bits;
{
    /* get the bits that are left and read the next word */
    unsigned temp =0;
        register int result = bitbuf;
    register int sbits = bits_left;
    bits -= bits_left;

    /* read next word of input */
    if (incnt > 2)

    /* this is the equivalent of the unused ReadWord function i added above */
        asm
        {
        move.l    inptr, A1;
        move.w    (A1)+, D1;
        rol.w    #8, D1;
        move.w    D1, bitbuf;
        move.l    A1, inptr;
        move.w    #16, bits_left;
        subq.l    #2, incnt;
        }
    else
    {
        bits_left = ReadByte(&bitbuf);
        bits_left += ReadByte(&temp);
        bitbuf |= (temp << 8);
        if (bits_left == 0)
        zipeof = 1;
    }

    /* get the remaining bits */
/*        result = result | (unsigned int) ((bitbuf & mask_bits[bits]) << sbits);
        bitbuf >>= bits;
        bits_left -= bits;
        return result;
*/
    /* The following is the equivalent of the above commented out
    section of code */
        asm
        {
        move.w    bits,    D2;
        lea        mask_bits, A0;
        asl.w    #1, D2;
        move.w    0(A0,D2.w),D0;
        move.w    bitbuf, D1;
        and.w    D1, D0;
        asl.w    sbits, D0;
        or.w    D0, result;
        lsr.w    bits, D1;
        move.w    D1, bitbuf;
        sub.w    bits, bits_left;
        }
        return result;
        
        

}

#define READBIT(nbits,zdest) { if (nbits <= bits_left) { zdest = (unsigned int)(bitbuf & mask_bits[nbits]); bitbuf >>= nbits; bits_left -= nbits; } else zdest = FillBitBuffer(nbits);}

/*
 * macro READBIT(nbits,zdest)
 *  {
 *      if (nbits <= bits_left) {
 *          zdest = (unsigned int)(bitbuf & mask_bits[nbits]);
 *          bitbuf >>= nbits;
 *          bits_left -= nbits;
 *      } else
 *          zdest = FillBitBuffer(nbits);
 *  }
 *
 */


/* ------------------------------------------------------------- */


/* #include "crc32.h" */
long crc32val;


/* ------------------------------------------------------------- */

void FlushOutput()
 /* flush contents of output buffer */
{
    UpdateCRC(outbuf, outcnt);
    Fwrite(outfd, outcnt, outbuf);
    outpos += outcnt;
    outcnt = 0;
    outptr = outbuf;
}

#define OUTB(intc) { *outptr++=intc; if (++outcnt==OUTBUFSIZ) FlushOutput(); }

/*
 *  macro OUTB(intc)
 *  {
 *      *outptr++=intc;
 *      if (++outcnt==OUTBUFSIZ)
 *          FlushOutput();
 *  }
 *
 */


/* ----------------------------------------------------------- */

void LoadFollowers()
{
        register int x;
        register int i;

    for (x = 255; x >= 0; x--) {
                READBIT(6,Slen[x]);
        for (i = 0; i < Slen[x]; i++) {
                        READBIT(8,followers[x][i]);
        }
    }
}


/* ----------------------------------------------------------- */
/*
 * The Reducing algorithm is actually a combination of two
 * distinct algorithms.  The first algorithm compresses repeated
 * byte sequences, and the second algorithm takes the compressed
 * stream from the first algorithm and applies a probabilistic
 * compression method.
 */

int L_table[] = {0, 0x7f, 0x3f, 0x1f, 0x0f};

int D_shift[] = {0, 0x07, 0x06, 0x05, 0x04};
int D_mask[]  = {0, 0x01, 0x03, 0x07, 0x0f};

int B_table[] = {8, 1, 1, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 5,
         5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 6, 6, 6,
         6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
         6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7,
         7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
         7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
         7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
         7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
         8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
         8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
         8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
  

     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
         8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
         8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
         8, 8, 8, 8};

/* ----------------------------------------------------------- */

void unReduce()
 /* expand probablisticly reduced data */
{
        register int lchar;
        int nchar;
        register int ExState;
        register int V;
        register int Len;

        factor = lrec.compression_method - 1;
    ExState = 0;
    lchar = 0;
    LoadFollowers();

        while (((outpos+outcnt) < lrec.uncompressed_size) && (!zipeof)) {
        if (Slen[lchar] == 0)
                        READBIT(8,nchar)      /* ; */
                else
        {
                        READBIT(1,nchar);
                        if (nchar != 0)
                                READBIT(8,nchar)      /* ; */
                        else
            {
                                int follower;
                                int bitsneeded = B_table[Slen[lchar]];
                                READBIT(bitsneeded,follower);
                                nchar = followers[lchar][follower];
            }
        }

        /* expand the resulting byte */
        switch (ExState) {

        case 0:
                        if (nchar != DLE)
                                OUTB(nchar) /*;*/
            else
                ExState = 1;
            break;

        case 1:
                        if (nchar != 0) {
                                V = nchar;
                Len = V & L_table[factor];
                if (Len == L_table[factor])
                    ExState = 2;
                else
                    ExState = 3;
            }
            else {
                                OUTB(DLE);
                ExState = 0;
            }
            break;

                case 2: {
                                Len += nchar;
                ExState = 3;
            }
            break;

                case 3: {
                register int i = Len + 3;
                int offset = (((V >> D_shift[factor]) &
                                D_mask[factor]) << 8) + nchar + 1;
                                longint op = (outpos+outcnt) - offset;

                /* special case- before start of file */
                while ((op < 0L) && (i > 0)) {
                    OUTB(0);
                    op++;
                    i--;
                }

                /* normal copy of data from output buffer */
                {
                    register int ix = (int) (op % OUTBUFSIZ);

                                        /* do a block memory copy if possible */
  

                                    if ( ((ix    +i) < OUTBUFSIZ) &&
                                             ((outcnt+i) < OUTBUFSIZ) ) {
                                                memcpy(outptr,&outbuf[ix],i);
                                                outptr += i;
                                                outcnt += i;
                                        }

                                        /* otherwise copy byte by byte */
                                        else while (i--) {
                                                OUTB(outbuf[ix]);
                                                if (++ix >= OUTBUFSIZ)
                                                        ix = 0;
                                        }
                                }

                ExState = 0;
            }
            break;
        }

                /* store character for next iteration */
                lchar = nchar;
        }
}


/* ------------------------------------------------------------- */
/*
 * Shrinking is a Dynamic Ziv-Lempel-Welch compression algorithm
 * with partial clearing.
 *
 */

void partial_clear()
{
        register int pr;
        register int cd;

    /* mark all nodes as potentially unused */
    for (cd = first_ent; cd < free_ent; cd++)
        prefix_of[cd] |= 0x8000;

    /* unmark those that are used by other nodes */
    for (cd = first_ent; cd < free_ent; cd++) {
        pr = prefix_of[cd] & 0x7fff;    /* reference to another node? */
                if (pr >= first_ent)            /* flag node as referenced */
            prefix_of[pr] &= 0x7fff;
    }

    /* clear the ones that are still marked */
    for (cd = first_ent; cd < free_ent; cd++)
        if ((prefix_of[cd] & 0x8000) != 0)
            prefix_of[cd] = -1;

    /* find first cleared node as next free_ent */
        cd = first_ent;
        while ((cd < maxcodemax) && (prefix_of[cd] != -1))
                cd++;
        free_ent = cd;
}


/* ------------------------------------------------------------- */

void unShrink()
{
#define  GetCode(dest) READBIT(codesize,dest)

    register int code;
    register int stackp;
    register int finchar;
    register int oldcode;
    register int incode;


    /* decompress the file */
    maxcodemax = 1 << max_bits;
    codesize = init_bits;
    maxcode = (1 << codesize) - 1;
    free_ent = first_ent;
    offset = 0;
    sizex = 0;

    for (code = maxcodemax; code > 255; code--)
        prefix_of[code] = -1;

    for (code = 255; code >= 0; code--) {
        prefix_of[code] = 0;
        suffix_of[code] = code;
    }

    GetCode(oldcode);
    if (zipeof)
        return;
    finchar = oldcode;

        OUTB(finchar);

        stackp = hsize;

    while (!zipeof) {
        GetCode(code);
        if (zipeof)
            return;

        while (code == clear) {
            GetCode(code);
            switch (code) {

            case 1:{
                    codesize++;
                    if (codesize == max_bits)
                        maxcode = maxcodemax;
                    else
                        maxcode = (1 << codesize) - 1;
                }
                break;

            case 2:
                partial_clear();
                break;
            }

            GetCode(code);
            if (zipeof)
                return;
        }


        /* special case for KwKwK string */
        incode = code;
        if (prefix_of[code] == -1) {
                        stack[--stackp] = finchar;
            code = oldcode;
        }


        /* generate output characters in reverse order */
        while (code >= first_ent) {
                        stack[--stackp] = suffix_of[code];
            code = prefix_of[code];
        }

        finchar = suffix_of[code];
                stack[--stackp] = finchar;


                /* and put them out in forward order, block copy */
                if ((hsize-stackp+outcnt) < OUTBUFSIZ) {
                        memcpy(outptr,&stack[stackp],hsize-stackp);
                        outptr += hsize-stackp;
                        outcnt += hsize-stackp;
                        stackp = hsize;
                }

                /* output byte by byte if we can't go by blocks */
                else while (stackp < hsize)
                        OUTB(stack[stackp++]);


        /* generate new entry */
        code = free_ent;
        if (code < maxcodemax) {
            prefix_of[code] = oldcode;
            suffix_of[code] = finchar;

            do
                code++;
            while ((code < maxcodemax) && (prefix_of[code] != -1));

            free_ent = code;
        }

        /* remember previous code */
        oldcode = incode;
    }

}


/* ------------------------------------------------------------- */ 
/*
 * Imploding
 * ---------
 *
 * The Imploding algorithm is actually a combination of two distinct
 * algorithms.  The first algorithm compresses repeated byte sequences
 * using a sliding dictionary.  The second algorithm is used to compress
 * the encoding of the sliding dictionary ouput, using multiple
 * Shannon-Fano trees.
 *
 */ 

   enum { maxSF        = 511 };

   typedef struct bintable { 
                 byte        Value; 
                 byte        BitLength;
                 word        zero;
                 word        one;
                 boolean    leaf;
              } bintable; 

   typedef struct sf_tree {   /* a shannon-fano  binary tree */ 
      bintable   bintree[maxSF];
      int        entries;
      int        MaxLength;
      int        Root;
   } sf_tree; 

   typedef sf_tree      *sf_treep; 

   sf_tree      lit_tree; 
   sf_tree      length_tree; 
   sf_tree      distance_tree; 
   boolean      lit_tree_present; 
   boolean      eightK_dictionary; 
   int          minimum_match_length;
   int          dict_bits;

#define EMPTY    0xffff
#define BRANCH   FALSE
#define LEAF     TRUE

void         SortLengths(tree)
register sf_tree *tree;

  /* Sort the Bit Lengths in ascending order, while retaining the order
    of the original lengths stored in the file */ 
{ 
   register int   x;
   register int   gap;
   bintable       t; 
   boolean        noswaps;
   register int   a, b;

   gap = tree->entries / 2; 

  do
  {
    do
    {
        noswaps = 1;
        for (x = 0; x <= (tree->entries - 1) - gap; x++) 
        {
            a = tree->bintree[x].BitLength;
            b = tree->bintree[x + gap].BitLength; 
            if ((a > b) || ((a == b) && (tree->bintree[x].Value > tree->bintree[x + gap].Value))) 
            { 
                t = tree->bintree[x]; 
                tree->bintree[x] = tree->bintree[x + gap]; 
                tree->bintree[x + gap] = t; 
                noswaps = 0;
            }
        }
    }  while (!noswaps);

    gap = gap / 2; 
   }  while (gap > 0);
} 


/* ----------------------------------------------------------- */ 

void         ReadLengths(tree)
register sf_tree *tree;

{ 
   register int          treeBytes;
   register int          i;
   register int          num, len;

  /* get number of bytes in compressed tree */
   READBIT(8,treeBytes);
   treeBytes++; 
   i = 0; 

   tree->MaxLength = 0;

 /* High 4 bits: Number of values at this bit length + 1. (1 - 16)
    Low  4 bits: Bit Length needed to represent value + 1. (1 - 16) */
   while (treeBytes > 0)
   {
      READBIT(4,len); len++;
      READBIT(4,num); num++;

      while (num > 0)
      {
         if (len > tree->MaxLength)
            tree->MaxLength = len;
         tree->bintree[i].BitLength = len;
         tree->bintree[i].Value = i;
         i++;
         num--;
      }

      treeBytes--;
   } 
} 


/* ----------------------------------------------------------- */ 



int BuildLevel( tree, Nodelevel, Nodelast )
register sf_tree    *tree;
register int        Nodelevel;
register int        Nodelast;

/* Build the next higher level of the binary tree */

{

register boolean isone = 0;
register int branchstop = Nodelevel;

for( Nodelast; Nodelast >= branchstop; Nodelast-- )
    {
        if( isone )
        {
            tree->bintree[Nodelevel].one  = Nodelast;
            isone--;
        }
        else
        {
            Nodelevel--;
            tree->bintree[Nodelevel].zero = Nodelast;
            tree->bintree[Nodelevel].one  = EMPTY;
            tree->bintree[Nodelevel].leaf = BRANCH;
            isone++;
        }
    }
    return Nodelevel;
}

/* ----------------------------------------------------------- */ 
        


void         BuildBinTree(tree)
register sf_tree *tree;
     /* Generate the Shannon-Fano binary trees */
/*
 * Works like this: we build up the binary tree from the lowest level
 * up to the root one level at a time.  BuildBinTree generates the
 * leaves for each level (-13-) and also builds the branches for these
 * leaves up one level (-12-).  Every time it hits the first leaf for
 * the next BitLength, it holds on to that leaf until the next iteration
 * and so does not enter it incorrectly into the present BitLength level.
 * Then it calls BuildLevel to the entire build existing tree up another
 * level (-11-), which coresponds to the level of the next set of
 * branches that BuildBinTree makes out of the next set of
 * BitLength-sized leaves, and then BuildLevel builds all of them up
 * another level, etc.  Then it finds the Root of the tree and put
 * it's location where ReadTree can find it.
 *
 */

{ 
    register word       Code;
    register int        CodeIncrement;
    int                 NowBitLength;
    register int        i;
    word                iCode;
    byte                zerodiscard;
    byte                branchspan;
    boolean             onezero;
    byte                LastBitLength;
    int                 Nodelast = 510;
    register int        last = 510;
    int                 ilast;
    byte                a;
    word                x;
    boolean             bitshift;


   Code = 0;
   iCode = 0;
   CodeIncrement = 0; 
   NowBitLength = 0;
   a = 0;
   bitshift = FALSE;

    i = tree->entries - 1;   /* either 255 or 63 */ 
    tree->bintree[i].leaf = LEAF;
    tree->bintree[last].zero = i;
    tree->bintree[last].leaf = BRANCH;
    tree->bintree[last].one = EMPTY;
    i--;
    NowBitLength = tree->bintree[i].BitLength;
    zerodiscard = (16 - NowBitLength); /* throw out unused trailing zero's */
    CodeIncrement = 1;

    for( i; i >= 0; i-- ) 
    {
    if( bitshift ) /* treat first leaf of this level special if it's
                      held over from previous level. */
    {
          onezero = x & 1;
          if( onezero && ((tree->bintree[last].one) != EMPTY))
          {
              last--;
              tree->bintree[last].zero = EMPTY;
              tree->bintree[last].leaf = BRANCH;
        }
        bitshift = FALSE;
        Code = x;
     }

     else

     {
         iCode += CodeIncrement;
          Code = iCode >> a; /* eat unused bits from all last levels */
     }

     onezero = Code & 1; /* find out if a Zero leaf or a One leaf */
     tree->bintree[i].leaf = LEAF;

      if( tree->bintree[i].BitLength != NowBitLength)
      {
          x = iCode >> (a + 1); /* save Code for the next level */
          bitshift = TRUE; /* remember to treat that first leaf as special */

        LastBitLength = NowBitLength;
        NowBitLength = tree->bintree[i].BitLength;
        branchspan= LastBitLength - NowBitLength;

            while (branchspan > 0) /* how many level up to build, in case */
                                   /* the next BitLength differs by > 1. */
        {
            ilast = last - 1;
            last = BuildLevel(tree, last, Nodelast );
            Nodelast = ilast;
            branchspan--;
        }

        a = (16 - (NowBitLength + zerodiscard)); /* eat trailing zero's */
        CodeIncrement = 1 << a;
        i++;
      }

      else

      {
      /* else build next level of branches */
          if ( onezero )
          {
              tree->bintree[last].one = i;
          }
          else
          {
              last--;
              tree->bintree[last].zero = i;
              tree->bintree[last].one = EMPTY;
              tree->bintree[last].leaf = BRANCH;
          }
      }

    }
    while( (ilast - last) > 0 ) /* find Root of tree */
    {
        ilast = last - 1;
        last = BuildLevel(tree, last, Nodelast );
        Nodelast = ilast;
    }

    tree->Root = last; /* store location of Root */
/*    printf("Built binary tree. Root = %d\n", tree->Root); */

}


/* ----------------------------------------------------------- */ 

/*    the old ReverseBit function went here    */

/* ----------------------------------------------------------- */ 

void         LoadTree(tree, treesize)
register sf_tree *tree;
register int treesize;

     /* allocate and load a shannon-fano tree from the compressed file */ 
{ 
   tree->entries = treesize; 
   ReadLengths(tree); 
   SortLengths(tree); 
   BuildBinTree(tree); 
} 


/* ----------------------------------------------------------- */ 

void         LoadTrees()
{ 
   eightK_dictionary = ((lrec.general_purpose_bit_flag & 0x02) != 0);   /* bit 1 */
   lit_tree_present = ((lrec.general_purpose_bit_flag & 0x04) != 0);   /* bit 2 */

   if (eightK_dictionary) 
      dict_bits = 7;
   else 
      dict_bits = 6; 

   if (lit_tree_present) 
   { 
      minimum_match_length = 3; 
      LoadTree(&lit_tree,256); 
   } 
   else
      minimum_match_length = 2; 

   LoadTree(&length_tree,64); 
   LoadTree(&distance_tree,64); 
}


/* ----------------------------------------------------------- */ 

int         ReadTree(tree)
register sf_tree *tree;

     /* read next byte using a shannon-fano binary tree */ 
{ 
register int          cur;
register int          b;

/*   *dest = -1;    in case of error */ 
/*     we don't do errors, sorry    */

    cur = tree->Root;
    while (! (tree->bintree[cur].leaf) )
    {
        READBIT(1,b);

        if ( b )
            cur = tree->bintree[cur].one;
        else
            cur = tree->bintree[cur].zero;

    }

    return (tree->bintree[cur].Value);
}


/* ----------------------------------------------------------- */ 

void         unImplode()
     /* expand imploded data */ 

{ 
register int          lout;
         longint      op;
register int          Length;
register int          Distance;

   LoadTrees(); 

   while ((!zipeof) && ((outpos+outcnt) < lrec.uncompressed_size))
   { 
      READBIT(1,lout);

      if (lout != 0)   /* encoded data is literal data */ 
      { 
         if (lit_tree_present)  /* use Literal Shannon-Fano tree */
            lout = ReadTree(&lit_tree);
         else 
            READBIT(8,lout);

         OUTB(lout);
      } 
      else             /* encoded data is sliding dictionary match */
      {                
         READBIT(dict_bits,lout);
      

 Distance = lout; 

         lout = ReadTree(&distance_tree); 
         Distance |= (lout << dict_bits);
         /* using the Distance Shannon-Fano tree, read and decode the
            upper 6 bits of the Distance value */ 

        Length =  ReadTree(&length_tree); 
         /* using the Length Shannon-Fano tree, read and decode the
            Length value */

         Length += minimum_match_length; 
         if (Length == (63 + minimum_match_length)) 
         { 
            READBIT(8,lout);
            Length += lout; 
         } 

        /* move backwards Distance+1 bytes in the output stream, and copy
          Length characters from this position to the output stream.
          (if this position is before the start of the output stream,
          then assume that all the data before the start of the output
          stream is filled with zeros) */ 

         op = (outpos+outcnt) - Distance - 1L;

          /* special case- before start of file */
          while ((op < 0L) && (Length > 0)) {
                  OUTB(0);
                  op++;
                  Length--;
          }

          /* normal copy of data from output buffer */
          {
                  register int ix = (int) (op % OUTBUFSIZ);

                  /* do a block memory copy if possible */
                  if ( ((ix    +Length) < OUTBUFSIZ) &&
                       ((outcnt+Length) < OUTBUFSIZ) ) {
                          memcpy(outptr,&outbuf[ix],Length);
                          outptr += Length;
                          outcnt += Length;
                  }

                  /* otherwise copy byte by byte */
                  else while (Length--) {
                          OUTB(outbuf[ix]);
                          if (++ix >= OUTBUFSIZ)
                                  ix = 0;
                  }
         }
      } 
   } 
} 



/* ---------------------------------------------------------- */

void extract_member()
{
    word     b;

    bits_left = 0;
    bitbuf = 0;
    incnt = 0;
    outpos = 0L;
    outcnt = 0;
    outptr = outbuf;
    zipeof = 0;
    crc32val = 0xFFFFFFFFL;


    /* create the output file with READ and WRITE permissions */
    if (create_output_file())
        exit(1);

        switch (lrec.compression_method) {

    case 0:        /* stored */
        {
            printf(" Extracting: %-12s\n", filename);
            while (ReadByte(&b))
                OUTB(b);
        }
        break;

        case 1: {
            printf("UnShrinking: %-12s\n", filename);
            unShrink();
        }
        break;

    case 2:
    case 3:
    case 4:
        case 5: {
            printf("  Expanding: %-12s\n", filename);
            unReduce();
        }
        break;

        case 6: {
                        printf("  Exploding: %-12s\n", filename);
                        unImplode();
        }
        break;

        default:
        printf("Unknown compression method.\n");
    }


    /* write the last partial buffer, if any */
    if (outcnt > 0) {
        UpdateCRC(outbuf, outcnt);
        Fwrite(outfd, outcnt, outbuf);
    }

    /* set output file date and time */
    set_file_time();

    close(outfd);

    crc32val = -1 - crc32val;
        if (crc32val != lrec.crc32)
                printf(" Bad CRC %08lx  (should be %08lx)\n", lrec.crc32, crc32val);

}


/* ---------------------------------------------------------- */

void get_string(len, s)
register unsigned int len;
register char *s;
{
    register long len2 = len;
    Fread(zipfd, len2, s);
    s[len] = 0;
}


/* ---------------------------------------------------------- */

void process_local_file_header()
{
    long lrec2 = sizeof(lrec);
    Fread(zipfd, lrec2, &lrec);

#ifdef HIGH_LOW
    swap_bytes(&lrec.filename_length);
    swap_bytes(&lrec.extra_field_length);
    swap_lbytes(&lrec.compressed_size);
    swap_lbytes(&lrec.uncompressed_size);
    swap_bytes(&lrec.compression_method);

/*    The following four fields must also be swapped.    */
/*    The basic reason why the original source by Sam Smith wouldn't work */
/*    on the ST, once you fixed the MSDOS functions to GEMDOS ones, */
/*    is because these four fields weren't swapped into the */
/*    68000 memory organisation format. */

    swap_bytes(&lrec.general_purpose_bit_flag);
    swap_lbytes(&lrec.crc32);
    swap_bytes(&lrec.last_mod_file_time);
    swap_bytes(&lrec.last_mod_file_date);

#endif

    get_string(lrec.filename_length, filename);
    get_string(lrec.extra_field_length, extra);
    extract_member();
}


/* ---------------------------------------------------------- */

void process_central_file_header()
{
    central_directory_file_header rec;
    char filename[STRSIZ];
    char extra[STRSIZ];
    char comment[STRSIZ];
    long rec2 = sizeof(rec);

    Fread(zipfd, rec2, &rec);

#ifdef HIGH_LOW
    swap_bytes(&rec.filename_length);
    swap_bytes(&rec.extra_field_length);
    swap_bytes(&rec.file_comment_length);
#endif

        get_string(rec.filename_length, filename);
    get_string(rec.extra_field_length, extra);
    get_string(rec.file_comment_length, comment);
}


/* ---------------------------------------------------------- */

void process_end_central_dir()
{
    end_central_dir_record rec;
    char comment[STRSIZ];
    long rec2 = sizeof(rec);

    Fread(zipfd, rec2, &rec);

#ifdef HIGH_LOW
    swap_bytes(&rec.zipfile_comment_length);
#endif

    get_string(rec.zipfile_comment_length, comment);
}


/* ---------------------------------------------------------- */

void process_headers()
{
    longint sig;
    long sig2 = sizeof(sig);

    while (1) {
        if (Fread(zipfd, sig2, &sig) != sizeof(sig))
            return;

#ifdef HIGH_LOW
        swap_lbytes(&sig);
#endif

                if (sig == LOCAL_FILE_HEADER_SIGNATURE)
          

process_local_file_header();
                else if (sig == CENTRAL_FILE_HEADER_SIGNATURE)
            process_central_file_header();
                else if (sig == END_CENTRAL_DIR_SIGNATURE) {
            process_end_central_dir();
            return;
        }
                else {
            printf("Invalid Zipfile Header\n");
            return;
        }
    }

}


/* ---------------------------------------------------------- */

void extract_zipfile()
{
    /*
     * open the zipfile for reading and in BINARY mode to prevent cr/lf
     * translation, which would corrupt the bitstreams 
     */

    if (open_input_file())
        exit(1);

    process_headers();

    close(zipfd);
}


/* ---------------------------------------------------------- */
/*
 * main program
 *
 */

void main(argc, argv)
int argc;
char **argv;
{
    if (argc != 2) {
        printf("\n%s\nCourtesy of:  S.H.Smith  and  The Tool Shop BBS,  (602) 279-2673.\n\n",VERSION);
        printf("You may copy and distribute this program freely, provided that:\n");
        printf("    1)   No fee is charged for such copying and distribution, and\n");
        printf("    2)   It is distributed ONLY in its original, unmodified state.\n\n");
        printf("If you wish to distribute a modified version of this program, you MUST\n");
        printf("include the source code.\n\n");
        printf("If you modify this program, I would appreciate a copy of the  new source\n");
        printf("code.   I am holding the copyright on the source code, so please don't\n");
        printf("delete my name from the program files or from the documentation.\n\n");
        printf("IN NO EVENT WILL I BE LIABLE TO YOU FOR ANY DAMAGES, INCLUDING ANY LOST\n");
        printf("PROFITS, LOST SAVINGS OR OTHER INCIDENTAL OR CONSEQUENTIAL DAMAGES\n");
        printf("ARISING OUT OF YOUR USE OR INABILITY TO USE THE PROGRAM, OR FOR ANY\n");
        printf("CLAIM BY ANY OTHER PARTY.\n\n");
        printf("Usage:  UnZip FILE[.zip]\n");
        exit(1);
    }

    /* .ZIP default if none provided by user */
    strcpy(zipfn, argv[1]);
    if (strchr(zipfn, '.') == NULL)
        strcat(zipfn, ".ZIP");

        /* allocate i/o buffers */
    inbuf = (byte *) (malloc(INBUFSIZ));
    outbuf = (byte *) (malloc(OUTBUFSIZ));
    if ((inbuf == NULL) || (outbuf == NULL)) {
        printf("Can't allocate buffers!\n");
        exit(1);
    }

        /* do the job... */
        extract_zipfile();
    exit(0);
}

